"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DiagnosticsProvider = void 0;
const tsyringe_1 = require("tsyringe");
const vscode_languageserver_1 = require("vscode-languageserver");
const vscode_uri_1 = require("vscode-uri");
const __1 = require("..");
const elmWorkspaceMatcher_1 = require("../../util/elmWorkspaceMatcher");
const noWorkspaceContainsError_1 = require("../../util/noWorkspaceContainsError");
const textDocumentEvents_1 = require("../../util/textDocumentEvents");
const elmMakeDiagnostics_1 = require("./elmMakeDiagnostics");
const typeInferenceDiagnostics_1 = require("./typeInferenceDiagnostics");
const astProvider_1 = require("../astProvider");
let DiagnosticsProvider = class DiagnosticsProvider {
    constructor() {
        this.elmAnalyseDiagnostics = null;
        this.settings = tsyringe_1.container.resolve("Settings");
        this.clientSettings = tsyringe_1.container.resolve("ClientSettings");
        if (this.clientSettings.elmAnalyseTrigger !== "never") {
            this.elmAnalyseDiagnostics = tsyringe_1.container.resolve(__1.ElmAnalyseDiagnostics);
        }
        this.elmMakeDiagnostics = tsyringe_1.container.resolve(elmMakeDiagnostics_1.ElmMakeDiagnostics);
        this.typeInferenceDiagnostics = tsyringe_1.container.resolve(typeInferenceDiagnostics_1.TypeInferenceDiagnostics);
        this.connection = tsyringe_1.container.resolve("Connection");
        this.events = tsyringe_1.container.resolve(textDocumentEvents_1.TextDocumentEvents);
        this.newElmAnalyseDiagnostics = this.newElmAnalyseDiagnostics.bind(this);
        this.elmWorkspaceMatcher = new elmWorkspaceMatcher_1.ElmWorkspaceMatcher((doc) => vscode_uri_1.URI.parse(doc.uri));
        this.workspaces = tsyringe_1.container.resolve("ElmWorkspaces");
        const astProvider = tsyringe_1.container.resolve(astProvider_1.ASTProvider);
        this.currentDiagnostics = {
            elmAnalyse: new Map(),
            elmMake: new Map(),
            elmTest: new Map(),
            typeInference: new Map(),
        };
        // register onChange listener if settings are not on-save only
        void this.settings.getClientSettings().then(({ elmAnalyseTrigger }) => {
            this.events.on("open", (d) => this.getDiagnostics(d, true, elmAnalyseTrigger));
            this.events.on("save", (d) => this.getDiagnostics(d, true, elmAnalyseTrigger));
            this.connection.onDidChangeWatchedFiles((event) => {
                const newDeleteEvents = event.changes
                    .filter((a) => a.type === vscode_languageserver_1.FileChangeType.Deleted)
                    .map((a) => a.uri);
                newDeleteEvents.forEach((uri) => {
                    this.currentDiagnostics.elmAnalyse.delete(uri);
                    this.currentDiagnostics.elmMake.delete(uri);
                    this.currentDiagnostics.elmTest.delete(uri);
                    this.currentDiagnostics.typeInference.delete(uri);
                });
                this.sendDiagnostics();
            });
            if (this.elmAnalyseDiagnostics) {
                this.elmAnalyseDiagnostics.on("new-diagnostics", this.newElmAnalyseDiagnostics);
            }
            if (elmAnalyseTrigger === "change") {
                this.events.on("change", (d) => this.getDiagnostics(d, false, elmAnalyseTrigger));
            }
            this.workspaces.forEach((workspace) => {
                workspace.getForest().treeIndex.forEach((treeContainer) => {
                    if (treeContainer.writeable) {
                        const treeDiagnostics = this.typeInferenceDiagnostics.createDiagnostics(treeContainer.tree, treeContainer.uri, workspace);
                        this.currentDiagnostics.typeInference.set(treeContainer.uri, treeDiagnostics);
                    }
                });
            });
            this.sendDiagnostics();
            astProvider.onTreeChange(({ uri, tree }) => {
                let workspace;
                try {
                    workspace = this.elmWorkspaceMatcher.getElmWorkspaceFor({ uri });
                }
                catch (error) {
                    if (error instanceof noWorkspaceContainsError_1.NoWorkspaceContainsError) {
                        this.connection.console.info(error.message);
                        return; // ignore file that doesn't correspond to a workspace
                    }
                    throw error;
                }
                this.currentDiagnostics.typeInference.set(uri, this.typeInferenceDiagnostics.createDiagnostics(tree, uri, workspace));
                this.sendDiagnostics();
            });
        });
    }
    newElmAnalyseDiagnostics(diagnostics) {
        this.currentDiagnostics.elmAnalyse = diagnostics;
        this.sendDiagnostics();
    }
    addOrMergeDiagnostics(map1, map2) {
        const result = new Map(map1);
        for (const key of map2.keys()) {
            const value = map2.get(key);
            if (value) {
                if (map1.has(key)) {
                    const value1 = map1.get(key);
                    if (value1) {
                        result.set(key, [...value, ...value1]);
                    }
                }
                else {
                    result.set(key, value);
                }
            }
        }
        return result;
    }
    sendDiagnostics() {
        let allDiagnostics = new Map();
        allDiagnostics = this.addOrMergeDiagnostics(allDiagnostics, this.currentDiagnostics.elmMake);
        allDiagnostics = this.addOrMergeDiagnostics(allDiagnostics, this.currentDiagnostics.elmTest);
        allDiagnostics = this.addOrMergeDiagnostics(allDiagnostics, this.currentDiagnostics.elmAnalyse);
        allDiagnostics = this.addOrMergeDiagnostics(allDiagnostics, this.currentDiagnostics.typeInference);
        for (const [uri, diagnostics] of allDiagnostics) {
            this.connection.sendDiagnostics({ uri, diagnostics });
        }
    }
    getDiagnostics({ document }, isSaveOrOpen, elmAnalyseTrigger) {
        return __awaiter(this, void 0, void 0, function* () {
            this.connection.console.info(`Diagnostics were requested due to a file ${isSaveOrOpen ? "open or save" : "change"}`);
            const uri = vscode_uri_1.URI.parse(document.uri);
            let workspace;
            try {
                workspace = this.elmWorkspaceMatcher.getElmWorkspaceFor(document);
            }
            catch (error) {
                if (error instanceof noWorkspaceContainsError_1.NoWorkspaceContainsError) {
                    this.connection.console.info(error.message);
                    return; // ignore file that doesn't correspond to a workspace
                }
                throw error;
            }
            const text = document.getText();
            if (isSaveOrOpen) {
                this.currentDiagnostics.elmMake = yield this.elmMakeDiagnostics.createDiagnostics(uri);
            }
            const elmMakeDiagnosticsForCurrentFile = this.currentDiagnostics.elmMake.get(uri.toString());
            if (this.elmAnalyseDiagnostics &&
                elmAnalyseTrigger !== "never" &&
                (!elmMakeDiagnosticsForCurrentFile ||
                    (elmMakeDiagnosticsForCurrentFile &&
                        elmMakeDiagnosticsForCurrentFile.length === 0))) {
                yield this.elmAnalyseDiagnostics.updateFile(uri, text);
            }
            this.sendDiagnostics();
        });
    }
};
DiagnosticsProvider = __decorate([
    tsyringe_1.injectable(),
    __metadata("design:paramtypes", [])
], DiagnosticsProvider);
exports.DiagnosticsProvider = DiagnosticsProvider;
//# sourceMappingURL=diagnosticsProvider.js.map